/*
 * de.superx.qa - a package for controlling QA routines
 * Copyright (C) 2023 Daniel Quathamer <danielq@memtext.de>
 *
 * This package is licensed under the CampusSource License; 
 * http://www.campussource.de/org/license/
 */
package de.superx.qa.bin;

import java.io.BufferedWriter;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.io.StringReader;
import java.io.StringWriter;
import java.net.URISyntaxException;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.sql.Statement;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Properties;
import java.util.StringTokenizer;

import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactoryConfigurationError;

import org.apache.log4j.Appender;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.apache.log4j.PropertyConfigurator;
import org.slf4j.LoggerFactory;

import org.jaxen.JaxenException;
import org.jaxen.XPath;
import org.jaxen.jdom.JDOMXPath;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.Namespace;
import org.jdom.input.SAXBuilder;
import org.springframework.mock.web.MockHttpServletRequest;
import org.springframework.mock.web.MockHttpServletResponse;
import org.springframework.mock.web.MockServletConfig;
import org.w3c.dom.NodeList;

import de.memtext.baseobjects.coll.NamedIdObjectList;
import de.memtext.util.LogUtils;
import de.superx.bin.Dosql;
import de.superx.bin.SxConnection;
import de.superx.common.FieldContainer;
import de.superx.common.Maske;
import de.superx.common.SuperX_el;
import de.superx.common.SxResultRow;
import de.superx.common.SxResultSet;
import de.superx.common.SxSqlHelper;
import de.superx.common.SxUser;
import de.superx.etl.EtlUtils;
import de.superx.qa.util.GetOpts;
import de.superx.sec.InputCheckRegistry;
import de.superx.servlet.SuperXManager;
import de.superx.servlet.SxPools;
import de.superx.servlet.UserInitializer;
import de.superx.util.PathAndFileUtils;
import de.superx.util.SqlStringUtils;
import net.sf.saxon.sxpath.XPathExpression;

public class QaTestcaseExecutor {
	private static String tcUniquename;
	private static int systeminfoId;
	private static String fieldSelCmdLine;
	private static int maskExecutionId;
	private static String maskeninfoId;
	private static String userName, userAdmin;
	private static int userId;
	private static long starttime ;
	private static long endtime ;
	private static String mandantenID="default";
	private static FieldContainer myFieldContainer;
	private static Hashtable<String, String> maskFields ;
	private static String resultXml;
	private static String resultHtml;
	private static String resultLog;
	private static String pathToDbProperties;
	private static Connection myConnection;
	private static String logFile;

	
	public QaTestcaseExecutor() {
			this.maskFields = new Hashtable();
			
			initLogging();
			resultHtml="";
	}

	private void init() throws SQLException, IOException, SecurityException, URISyntaxException
	{
		
		/*File f = new File(logConf);
		if (!f.exists()) {
			throw new IOException("Datei nicht gefunden: " + logConf);
		}
		FileInputStream ins = new FileInputStream(logConf);
		LogManager MyLogManager = java.util.logging.LogManager.getLogManager();
		MyLogManager.readConfiguration(ins);
		String level = MyLogManager.getProperty(".level");
		logger.info("Using Loggging-Level " + level);
		logger.info("Start Lauf");*/
		//initLogging(true, Level.INFO);
		//logger.info("START");
		try {
			myConnection = this.getConnection(pathToDbProperties);
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		if(myConnection==null)
			throw new SQLException("No connection possible ");

		
		//System.out.println("fieldSel:"+fieldSel);
		
		writeToLog("QaTestcaseExecutor inited");
		
	}
	
	private static Connection getConnection(String pathToDbProperties) throws Exception {
		String myDefaultPropFile=PathAndFileUtils.getWebinfDirectory()+PathAndFileUtils.PATHSEP+"db.properties";
		if(pathToDbProperties==null)
			pathToDbProperties=myDefaultPropFile;
		Connection myConnection;
			SxConnection mySxConnection = null;
			mySxConnection = new SxConnection();
			mySxConnection.setPropfile(pathToDbProperties);

			myConnection = mySxConnection.getConnection();


		
		return myConnection;
	}
	
	public static void main(String[] args) {
		String usage="usage: java de.superx.qa.bin.QaTestcaseExecutor -userid:1(optional) -tc:abc(optional) -systeminfo_id:9(optional)" ;
		GetOpts.setOpts(args);
		
		int returnCode=0;
		int nrOfTestcases=0;
		int nrOfSuccessfulTestcases=0;
		if (GetOpts.isPresent("-userid"))
		userId = Integer.valueOf(GetOpts.getValue("-userid"));
		
		if (GetOpts.isPresent("-tc"))
				tcUniquename= GetOpts.getValue("-tc");
		if (GetOpts.isPresent("-systeminfo_id"))
				systeminfoId= Integer.valueOf(GetOpts.getValue("-systeminfo_id"));
		
	
		QaTestcaseExecutor myExec=new QaTestcaseExecutor(); //TODO:Mandantid
		
		try {
			myExec.init();
			int[] myTestcases=getTestcases();
			nrOfTestcases=myTestcases.length;
			if(nrOfTestcases>0)
			{
				writeToLog("nrOfTestcases: "+nrOfTestcases);
				for (int tcNr = 0;tcNr < nrOfTestcases; tcNr++) 
				{
					returnCode=executeTestcase(myTestcases[tcNr]);
					//reset static variables:
					resultLog="";
					resultXml="";
					resultHtml="";
					tcUniquename="" ;
					if(returnCode==0)
						nrOfSuccessfulTestcases++;
				}
			}
			myConnection.close();
			
		} catch (Exception e) {
			// TODO Auto-generated catch block

			e.printStackTrace();
			System.err.println(e.toString());
			System.exit(1);
		}
		System.out.println("Executed testcases:"+ nrOfTestcases+"; successful: "+nrOfSuccessfulTestcases);
		
		
		System.exit(returnCode);	

	}
	
	private static int[] getTestcases() throws Exception
	{
		//TODO_ Umlaute erlauben:
		//if(!SqlStringUtils.checkValidKeyEntry(tcUniquename) || !SqlStringUtils.checkValidKeyEntry(systeminfoId))
		//	throw new Exception("Invalid Testcase "+tcUniquename);

		String sql = ""
				+ "SELECT \n"
				+ "E.tid as mask_execution_id\n"
				+ "FROM qa_mask_execution E left outer join userinfo U\n"
				+ " on (U.tid=E.userinfo_id)"
				+ "where E.is_active=1";
		if(tcUniquename!=null)
			sql +=" and E.uniquename ='"+tcUniquename+"'";
		if(systeminfoId>0)
			sql +=" and E.systeminfo_id="+systeminfoId;
		if(userId>0)
			sql +=" and E.userinfo_id="+userId;
		sql +=" order by 1;";
		SuperX_el el = new SuperX_el();
		SxSqlHelper sh=new SxSqlHelper();

		sh.execute(sql, myConnection, el);
		if (el.getError_String() != null
				&& !el.getError_String().trim().equals(""))
			throw new SQLException("\nProblem bei Testfall Sammlung DETAILS:" + "\n\n  Meldung:"
					+ el.getError_String() + "\n  sql:" + sql);
		SxResultSet result=  el.getResultSet();
		int nrOfTestcases=result.size();
		int[] testCases=new int[nrOfTestcases];

		int rownr=0;
		for (Iterator it = result.iterator(); it.hasNext();) {
			SxResultRow row = (SxResultRow) it.next();
			maskExecutionId=(int) row.get(0);
			testCases[rownr]=maskExecutionId;
			rownr++;
		}
		
		return testCases;

	}
	private static int executeTestcase(int maskExecutionId) throws Exception
	{
		
		int returnCode;
		starttime = new java.util.Date().getTime() ;
		String fieldSel="";
		String felderinfoId="";
		String felderinfoName="";
		String tcName="";
		String fieldValue="";
		String sql = ""
				+ "SELECT \n"
				+ "E.maskeninfo_id,"
				+ "E.userinfo_id as userid,"
				+ "U.benutzer as username,"
				+ "S.mask_execution_id,\n"
				+ "S.felderinfo_id,\n"
				+ "trim(F.name) as felderinfo_name,"
				+ "string_not_null(S.field_value) as field_value, \n"
				+ "U.administration,  \n"
				+ "E.name, \n"
				+ "E.uniquename \n"
				+ "FROM felderinfo F,qa_mask_field_sel S, qa_mask_execution E left outer join userinfo U\n"
				+ " on (U.tid=E.userinfo_id)"
				+ "where F.tid=S.felderinfo_id "
				+ "and E.tid=S.mask_execution_id\n"
				+ "and E.tid="+maskExecutionId+";";
		SuperX_el el = new SuperX_el();
		SxSqlHelper sh=new SxSqlHelper();

		sh.execute(sql, myConnection, el);
		if (el.getError_String() != null
				&& !el.getError_String().trim().equals(""))
			throw new SQLException("\nProblem bei Testfall DETAILS:" + "\n\n  Meldung:"
					+ el.getError_String() + "\n  sql:" + sql);
		SxResultSet result=  el.getResultSet();
		int rownr=0;
		for (Iterator it = result.iterator(); it.hasNext();) {
			rownr++;
			SxResultRow row = (SxResultRow) it.next();
			maskeninfoId=row.get(0).toString().trim();
			userId=(int) row.get(1);
			userName=row.get(2).toString().trim();
			maskExecutionId=(int) row.get(3);
			felderinfoId=row.get(4).toString().trim();
			felderinfoName=row.get(5).toString().trim();
			fieldValue=row.get(6).toString().trim();
			userAdmin=row.get(7).toString().trim();
			tcName=row.get(8).toString().trim();
			tcUniquename=row.get(9).toString().trim();
			maskFields.put(felderinfoName,fieldValue);
		}
		writeToLog("Testcase "+tcName + "(Schlüssel="+tcUniquename+" ID="+ maskExecutionId+")  initialized");
		returnCode=executeMask();
		endtime = new java.util.Date().getTime() ;
		writeToLog("Mask "+maskeninfoId +" executed in "+(endtime-starttime)/1000+" sec. with return Code "+returnCode);
		int maskExecutionResultId=saveExecutionResult(maskExecutionId,returnCode,resultLog);
		if(returnCode ==0)
		{
			returnCode=checkAssertions(maskExecutionId,maskExecutionResultId);
			
		}
		else
			writeToLog("Testcase error in masc execution "+maskExecutionId);
		return returnCode;
	}
	private static int executeMask()
	{	
	 int returnCode=0;
	 MockHttpServletRequest mock ;
	 MockHttpServletResponse mockResponse = new MockHttpServletResponse();
	 MockServletConfig mockServletConfig = new MockServletConfig();
	 SxUser user=null;
	 mock = new MockHttpServletRequest();
	 Enumeration fieldEnum = maskFields.keys();
	 while (fieldEnum.hasMoreElements()) {
		 String paramName = (String) fieldEnum.nextElement();
		 String paramVal= maskFields.get(paramName).toString();
		 mock.addParameter(paramName, paramVal);
			
	 }
		try {
			SuperXManager.initKettleEnv();
			SxPools.init();
			SxPools.get(mandantenID).init();
			SxPools.get(mandantenID).initLogging(true);
			SxPools.resetAllPools();
			
			user = UserInitializer.initUser(mandantenID, userName, userId, userAdmin);
			
		} catch (Exception e) {
			writeToLog("Fehler beim Aufbau der Connection: " + e.toString());
			return 1;
		}
		Locale desiredLocale = new Locale(de.superx.util.SqlStringUtils.getEncoding());
		SuperXManager.maxRows = 10000;
		Maske maske = null;
		InputCheckRegistry.registerDefaultChecks();
		try {
			maske = new Maske(mandantenID, user, new Integer(maskeninfoId), desiredLocale);
			/*NamedIdObjectList fields = maske.readFelderFromDb(user);
			myFieldContainer = maske.getIndividualFields();
			myFieldContainer.addAll(fields);*/
			maske.setFieldDefaults(user, mock, true);
			resultXml=getResultXml(mock, mockResponse, maske, user, desiredLocale);
			if(resultXml==null)
			{
				writeToLog("resultXml  für Maske " + maskeninfoId + " konnte nicht erzeugt werden");
			}
			else
			{
			writeToLog("resultXml Länge " + resultXml.length() + " für Maske " + maskeninfoId + " erzeugt");
			transformResultXmltoHtml();
			}
			
		} catch (Exception e) {
			e.printStackTrace();
			resultLog="Fehler beim Ausführen der Maske " + maskeninfoId + ": " + e.toString();
			returnCode=1;
		}
		user=null;
	return returnCode;
	}
	private static  int checkAssertions(int execId,int maskExecutionResultId) throws Exception
	{
		int resultCode=0;
		int assertionId,assertionRownr,assertionColnr;
		Double minVal,maxVal,foundVal=null;
		String caption;
		String sql = "select "
			+"tid,"
			+"rownr,"
			+"colnr,"
			+"result_value_min,"
			+"result_value_max,"
			+ "caption"
			+" from qa_mask_execution_assert"
			+ " where is_active=1"
			+" and mask_execution_id="+execId+";";
		
		
		/*try {
			myConnection = getConnection(pathToDbProperties);
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}*/
		if(myConnection==null)
			throw new SQLException("No connection possible ");
		SuperX_el el = new SuperX_el();
		SxSqlHelper sh=new SxSqlHelper();

		sh.execute(sql, myConnection, el);
		if (el.getError_String() != null
				&& !el.getError_String().trim().equals(""))
			throw new SQLException("\nProblem bei Testfall Assertion DETAILS:" + "\n\n  Meldung:"
					+ el.getError_String() + "\n  sql:" + sql);
		SxResultSet result=  el.getResultSet();
		int rownr=0;
		for (Iterator it = result.iterator(); it.hasNext();) {
			rownr++;
			SxResultRow row = (SxResultRow) it.next();
			assertionId=(Integer) row.get(0);
			assertionRownr=(Integer) row.get(1);
			assertionColnr=(Integer) row.get(2);
			minVal=(Double) row.get(3);
			maxVal=(Double) row.get(4);
			caption=(String) row.get(5).toString();
			try {
				foundVal=getValFromXml(assertionRownr,assertionColnr);
			} catch (Exception e) {
				writeToLog("Testfall ID "  +execId+" mit Maskennr. "+maskeninfoId+ " Assertion-ID "+assertionId+ " " + caption+ " kein Ergebnis:\n" +e.toString());
				resultCode=1;
				return resultCode;
			}
			if(foundVal != null && foundVal >=minVal && foundVal<=maxVal)
				{
				writeToLog("Testfall ID "  +execId+" mit Maskennr. "+maskeninfoId+ " Assertion-ID "+assertionId+ " " + caption+ " und Wert "+ foundVal+" erfolgreich");
				}
			else
			{
				resultCode=1;
				writeToLog("Testfall ID "  +execId+" mit Maskennr. "+maskeninfoId+ " Assertion-ID "+assertionId+ " " + caption +" und Wert "+ foundVal+ " beendet mit Warnung/Fehler");
			}
			saveAssertResult(assertionId,maskExecutionResultId,foundVal,resultCode,minVal,maxVal);
		}
		
		//System.out.println("fieldSel:"+fieldSel);
		//myConnection.close();
		return resultCode;
	}
	private static void saveAssertResult(int maskExecutionAssertId,int maskExecutionResultId,Double resultValue,int resultCode,double minVal,double maxVal)  throws SQLException
	{	
		String sql="insert into qa_mask_execution_assert_result(mask_execution_assert_id, mask_execution_result_id, result_code, assert_result_value_min, assert_result_value_max";
		if(resultValue==null)
			sql+=") values(?,?,?,?,?)";
		else
			sql +=",result_value) values(?,?,?,?,?,?);";
		PreparedStatement st=myConnection.prepareStatement(sql);
		st.setInt(1,maskExecutionAssertId);
		st.setInt(2,maskExecutionResultId);
		st.setInt(3,resultCode);
		st.setDouble(4, minVal);
		st.setDouble(5, maxVal);
		if(resultValue!=null)
			st.setDouble(6, resultValue);

		boolean executed= st.execute();
	
	}
	private static int saveExecutionResult(int execId, int maskExecutionResult,String log ) throws SQLException
	{	
		int ret=0;
		java.sql.Timestamp startTimeDate = new java.sql.Timestamp(starttime);
		java.sql.Timestamp endTimeDate = new java.sql.Timestamp(endtime);
		String sql="insert into qa_mask_execution_result(mask_execution_id, execution_start, execution_end, result_code,result_log,result_xml,result_html)";
		sql +="values(?,?,?,?,?,?,?);";
		PreparedStatement st=myConnection.prepareStatement(sql);
		st.setInt(1,execId);
		st.setTimestamp(2, startTimeDate);
		st.setTimestamp(3, endTimeDate);
		st.setInt(4,maskExecutionResult);
		st.setString(5,log);
			//im Fehler/Warnungsfall Ergebnis speichern:
		st.setString(6,resultXml);
		st.setString(7,resultHtml);
	
		boolean executed= st.execute();
		//get generated tid:
		sql="select currval('qa_mask_execution_result_tid_seq')::integer;";
		SuperX_el el = new SuperX_el();
		SxSqlHelper sh=new SxSqlHelper();
		sh.execute(sql, myConnection, el);
		SxResultSet result=  el.getResultSet();
		int rownr=0;
		for (Iterator it = result.iterator(); it.hasNext();) {
			rownr++;
			SxResultRow row = (SxResultRow) it.next();
			ret=(Integer) row.get(0);
		}
			
		return ret;
		
	}	private static Double getValFromXml(int assertionRownr,int assertionColnr) throws Exception
	{
		Double foundval = null;
		SAXBuilder builder = new SAXBuilder();
		Document doc = builder.build(new StringReader(resultXml));
		Namespace ns = doc.getRootElement().getNamespace();
	    String xpathStr = "/ergebnisse/ergebnis/ergebniselement/sqlerg/row[@no="+(assertionRownr-1)+"]/col[@id="+(assertionColnr-1)+"]/wert";
	    XPath expression = new JDOMXPath(xpathStr);
	    Element myNode = (Element) expression.selectSingleNode(doc);
 	    if(myNode==null)
	    	return null;
	    else
	    {
	    	String myNodeVal=myNode.getTextNormalize();
	    	return foundval.parseDouble(myNodeVal);
	    }
		
	}
	private static String getResultXml(MockHttpServletRequest mock_par, MockHttpServletResponse mockResponse_par,
			Maske maske, SxUser user_par,  Locale desiredLocale) {
		String currentXml_local = null;
		maske.setMaxOffset(10000000);
		maske.setReuseResult(false);
		maske.resetTableStylesheet();
		maske.setSelectedTableStylesheetFileAndContenttype("tabelle_xml.xsl");
		maske.setDesiredContenttype("text/xml");
		try {
			currentXml_local = maske.runQuery(user_par, mock_par, null).toString();
			resultLog=("Ausgabe Maskenprotokoll\n" + SuperXManager.activityLog.toString());
			System.out.println("Maskenergebnis wird lokalisiert");
			currentXml_local = SxPools.get(mandantenID).localize(currentXml_local, desiredLocale);

		} catch (Exception e) {
			resultLog="Fehler beim Erstellen des XML für Maske " + e.toString();
			System.err.println("Fehler beim Erstellen des XML für Maske " + e.toString());
		}
		return currentXml_local;

	}
	public void initLogging()   {
		File f;
		try {
			logFile=PathAndFileUtils.getWebinfPath()+File.separator+"logs"+File.separator+"superx_qa.log" ;
			f = new File(logFile);
			if(f.exists())
					f.delete();
			
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
				}
		writeToLog("Start");
		
	}
	
	private static void writeToLog(String msg) 
	{
		FileWriter f;
		try {
			f = new FileWriter(logFile,true);
			long jetzt= new java.util.Date().getTime() ;
			Date date = new Date();
			date.setTime(jetzt);
			SimpleDateFormat format = new SimpleDateFormat("yyyy-MM-dd hh:mm:ss");
			f.write(format.format(date)+": " +msg+"\n");
			f.close();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			System.err.println(msg);
		}
		
	}
	private static void transformResultXmltoHtml()
	{
		String xslFile="";
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		PrintStream myStream = new PrintStream(baos);
		try {
			StringWriter myOutwriter= new StringWriter();
			xslFile=PathAndFileUtils.getWebinfDirectory()+PathAndFileUtils.PATHSEP+".."+PathAndFileUtils.PATHSEP+"xml"+PathAndFileUtils.PATHSEP+"tabelle_html_qa.xsl";
			File f=new File(xslFile);
			//in development mode the XSL is not available:
			if(!f.exists())
				xslFile=PathAndFileUtils.getWebinfDirectory()+PathAndFileUtils.PATHSEP+".."+PathAndFileUtils.PATHSEP+"xml"+PathAndFileUtils.PATHSEP+"tabelle_html2.xsl";
				
			javax.xml.transform.TransformerFactory tFactory = javax.xml.transform.TransformerFactory.newInstance("net.sf.saxon.TransformerFactoryImpl",null);
			javax.xml.transform.Transformer transformer = tFactory.newTransformer(new javax.xml.transform.stream.StreamSource(xslFile));
			StringReader s1 = new StringReader(resultXml);
			transformer.setOutputProperty(javax.xml.transform.OutputKeys.ENCODING, SqlStringUtils.getEncoding());
			transformer.setOutputProperty(javax.xml.transform.OutputKeys.METHOD, "html");
			/* TODO: hier Parameter einzelne Zeile übergeben:
			 *                String paramname = param.substring(0, param.indexOf("="));
                    String paramvalue = param.substring(param.indexOf("=") + 1, param.length());
                    transformer.setParameter(paramname, paramvalue);
     
			 */
			 transformer.transform(new javax.xml.transform.stream.StreamSource(s1), new javax.xml.transform.stream.StreamResult(myOutwriter));
			 resultHtml=myOutwriter.toString();
			 myOutwriter.close();
			 s1.close();
			 
		} catch (Exception e) {
				writeToLog("resultXml konnte nicht zu HTML konvertiert werden:\n"+ e.toString());
				resultLog+=("resultXml konnte nicht zu HTML konvertiert werden:\n"+ e.toString());
				resultHtml="resultXml konnte nicht zu HTML konvertiert werden:" + resultXml;
		}
	}
}



